# flake8: noqa

"""
Linear Algebra Routines (deprecated)

.. warning::
    This module is deprecated and will be replaced by `pylinalg
    <https://github.com/pygfx/pylinalg>`_ in a future version.

"""

from .utils import *
from .cylindrical import *
from .euler import *
from .matrix3 import *
from .matrix4 import *
from .quaternion import *
from .spherical import *
from .vector3 import *
from .vector4 import *
